import type { ArkErrors, BaseRoot, Disjoint, JsonSchema, MetaSchema, Morph, UndeclaredKeyBehavior } from "@ark/schema";
import type { anyOrNever, array, Callable, ErrorMessage, inferred, Json, unset } from "@ark/util";
import type { inferIntersection } from "../intersect.ts";
import type { type } from "../keywords/ark.ts";
import type { applyConstraint, Default, distill, inferMorphOut, inferPipes, Optional } from "../keywords/ast.ts";
import type { Scope } from "../scope.ts";
import type { ArrayType } from "./array.ts";
import type { instantiateType } from "./instantiate.ts";
/** @ts-ignore cast variance */
interface Type<out t = unknown, $ = {}> extends Callable<(data: unknown) => distill.Out<t> | ArkErrors> {
    t: t;
    inferBrandableIn: distill.brandable.In<t>;
    inferBrandableOut: distill.brandable.Out<t>;
    inferIntrospectableOut: distill.introspectable.Out<t>;
    infer: distill.Out<t>;
    inferIn: distill.In<t>;
    [inferred]: t;
    json: Json;
    toJsonSchema(): JsonSchema;
    description: string;
    expression: string;
    internal: BaseRoot;
    $: Scope<$>;
    assert(data: unknown): this["infer"];
    allows(data: unknown): data is this["inferIn"];
    traverse(data: unknown): distill.Out<t> | ArkErrors;
    configure(meta: MetaSchema): this;
    describe(description: string): this;
    optional<r = applyConstraint<t, Optional>>(): instantiateType<r, $>;
    default<value extends this["infer"], r = (In?: this["inferBrandableIn"]) => Default<value>>(value: value): instantiateType<r, $>;
    onUndeclaredKey(behavior: UndeclaredKeyBehavior): this;
    onDeepUndeclaredKey(behavior: UndeclaredKeyBehavior): this;
    from(literal: this["inferIn"]): this["infer"];
    as<t = unset>(...args: validateChainedAsArgs<t>): instantiateType<t, $>;
    get in(): instantiateType<this["inferBrandableIn"], $>;
    get out(): instantiateType<this["inferIntrospectableOut"], $>;
    intersect<const def, r = type.infer<def, $>>(def: type.validate<def, $>): instantiateType<inferIntersection<t, r>, $> | Disjoint;
    and<const def, r = type.infer<def, $>>(def: type.validate<def, $>): instantiateType<inferIntersection<t, r>, $>;
    or<const def, r = type.infer<def, $>>(def: type.validate<def, $>): instantiateType<t | r, $>;
    array(): ArrayType<t[], $>;
    pipe: ChainedPipes<t, $>;
    equals<def>(def: type.validate<def, $>): this is instantiateType<type.infer<def, $>, $>;
    extract<def>(r: type.validate<def, $>): instantiateType<Extract<t, type.infer<def, $>>, $>;
    exclude<def>(r: type.validate<def, $>): instantiateType<Exclude<t, type.infer<def, $>>, $>;
    extends<def>(other: type.validate<def, $>): this is instantiateType<type.infer<def, $>, $>;
    overlaps<def>(r: type.validate<def, $>): boolean;
    distribute<mapOut, reduceOut = mapOut[]>(mapBranch: (branch: Type, i: number, branches: array<Type>) => mapOut, reduceMapped?: (mappedBranches: mapOut[]) => reduceOut): reduceOut;
    /** @deprecated */
    apply: Function["apply"];
    /** @deprecated */
    bind: Function["bind"];
    /** @deprecated */
    call: Function["call"];
    /** @deprecated */
    caller: Function;
    /** @deprecated */
    length: number;
    /** @deprecated */
    name: string;
    /** @deprecated */
    prototype: Function["prototype"];
    /** @deprecated */
    arguments: Function["arguments"];
    /** @deprecated */
    Symbol: never;
}
interface ChainedPipeSignature<t, $> {
    <a extends Morph<distill.Out<t>>, r = instantiateType<inferPipes<t, [a]>, $>>(a: a): r;
    <a extends Morph<distill.Out<t>>, b extends Morph<inferMorphOut<a>>, r = instantiateType<inferPipes<t, [a, b]>, $>>(a: a, b: b): r;
    <a extends Morph<distill.Out<t>>, b extends Morph<inferMorphOut<a>>, c extends Morph<inferMorphOut<b>>, r = instantiateType<inferPipes<t, [a, b, c]>, $>>(a: a, b: b, c: c): r;
    <a extends Morph<distill.Out<t>>, b extends Morph<inferMorphOut<a>>, c extends Morph<inferMorphOut<b>>, d extends Morph<inferMorphOut<c>>, r = instantiateType<inferPipes<t, [a, b, c, d]>, $>>(a: a, b: b, c: c, d: d): r;
    <a extends Morph<distill.Out<t>>, b extends Morph<inferMorphOut<a>>, c extends Morph<inferMorphOut<b>>, d extends Morph<inferMorphOut<c>>, e extends Morph<inferMorphOut<d>>, r = instantiateType<inferPipes<t, [a, b, c, d, e]>, $>>(a: a, b: b, c: c, d: d, e: e): r;
    <a extends Morph<distill.Out<t>>, b extends Morph<inferMorphOut<a>>, c extends Morph<inferMorphOut<b>>, d extends Morph<inferMorphOut<c>>, e extends Morph<inferMorphOut<d>>, f extends Morph<inferMorphOut<e>>, r = instantiateType<inferPipes<t, [a, b, c, d, e, f]>, $>>(a: a, b: b, c: c, d: d, e: e, f: f): r;
    <a extends Morph<distill.Out<t>>, b extends Morph<inferMorphOut<a>>, c extends Morph<inferMorphOut<b>>, d extends Morph<inferMorphOut<c>>, e extends Morph<inferMorphOut<d>>, f extends Morph<inferMorphOut<e>>, g extends Morph<inferMorphOut<f>>, r = instantiateType<inferPipes<t, [a, b, c, d, e, f, g]>, $>>(a: a, b: b, c: c, d: d, e: e, f: f, g: g): r;
}
export interface ChainedPipes<t, $> extends ChainedPipeSignature<t, $> {
    try: ChainedPipeSignature<t, $>;
}
type validateChainedAsArgs<t> = [
    t
] extends [unset] ? [
    t
] extends [anyOrNever] ? [
] : [
    ErrorMessage<"as requires an explicit type parameter like myType.as<t>()">
] : [];
export type { Type as BaseType };
